//----------------------------------------------------------------------------
// Moving obstacle simulator
// Christopher Rasmussen
// (c) University of Delaware, 2012
//----------------------------------------------------------------------------

#include "ros/ros.h"
#include "std_msgs/String.h"
#include "sensor_msgs/LaserScan.h"
#include "visualization_msgs/Marker.h"
#include "geometry_msgs/Twist.h"
#include "move_base_msgs/MoveBaseAction.h"
#include "nav_msgs/Odometry.h"
#include "nav_msgs/Path.h"

#include <gazebo_msgs/SetModelState.h>
#include <gazebo_msgs/SpawnModel.h>
#include <gazebo_msgs/DeleteModel.h>

#include <LinearMath/btQuaternion.h> // Needed to convert rotation ...
#include <LinearMath/btMatrix3x3.h>  // ... quaternion into Euler angles

#include <pcl/point_types.h>

#include <tf/transform_listener.h>

//----------------------------------------------------------------------------

using namespace std;

ros::NodeHandle *nhp;

int num_objects = 2;

vector <string> object_model_name;
vector <double> object_x;
vector <double> object_y;
vector <double> object_z;

ros::ServiceClient model_state_client;
gazebo_msgs::SetModelState setmodelstate;
gazebo_msgs::ModelState modelstate;

ros::ServiceClient spawn_model_client;
gazebo_msgs::SpawnModelRequest spawnmodelrequest;
gazebo_msgs::SpawnModel spawnmodel;

ros::ServiceClient delete_model_client;
gazebo_msgs::DeleteModelRequest deletemodelrequest;
gazebo_msgs::DeleteModel deletemodel;

//----------------------------------------------------------------------------

string box_urdf = "<robot name=\"big_box\"> \
  <link name=\"my_box\"> \
    <inertial> \
      <origin xyz=\"0 0 0\" /> \
      <mass value=\"0.1\" /> \
      <inertia  ixx=\"0.1\" ixy=\"0.0\"  ixz=\"0.0\"  iyy=\"100.0\"  iyz=\"0.0\"  izz=\"0.1\" /> \
    </inertial> \
    <visual> \
      <origin xyz=\"0 0 0\"/> \
      <geometry> \
        <box size=\"1 1 1\" /> \
      </geometry> \
    </visual> \
    <collision> \
      <origin xyz=\"0 0 0\"/> \
      <geometry> \
        <box size=\"1 1 1\" /> \
      </geometry> \
    </collision> \
  </link> \
  <gazebo reference=\"my_box\"> \
    <material>Gazebo/Red</material> \
    <turnGravityOff>true</turnGravityOff> \
  </gazebo> \
</robot>"; 

//----------------------------------------------------------------------------

string ball_urdf = "<robot name=\"small_ball\"> \
  <link name=\"my_ball\"> \
    <inertial> \
      <origin xyz=\"0 0 0\" /> \
      <mass value=\"100.0\" /> \
      <inertia  ixx=\"10\" ixy=\"0.0\"  ixz=\"0.0\"  iyy=\"100.0\"  iyz=\"0.0\"  izz=\"10\" /> \
    </inertial> \
    <visual> \
      <origin xyz=\"0 0 0\"/> \
      <geometry> \
        <sphere radius=\"0.25\" /> \
      </geometry> \
    </visual> \
    <collision> \
      <origin xyz=\"0 0 0\"/> \
      <geometry> \
        <sphere radius=\"0.25\" /> \
      </geometry> \
    </collision> \
  </link> \
  <gazebo reference=\"my_ball\"> \
    <material>Gazebo/Blue</material> \
    <turnGravityOff>true</turnGravityOff> \
  </gazebo> \
</robot>"; 

//----------------------------------------------------------------------------

// initialize service client and set the unchanging part of the modelstate message

void initialize_modelstates()
{
  int i;

  model_state_client = nhp->serviceClient<gazebo_msgs::SetModelState>("/gazebo/set_model_state");

  modelstate.reference_frame = "map";

  modelstate.pose.orientation.x = 0;
  modelstate.pose.orientation.x = 0;
  modelstate.pose.orientation.x = 0;
  modelstate.pose.orientation.w = 1;

  modelstate.twist.linear.x = 0;
  modelstate.twist.linear.y = 0;
  modelstate.twist.linear.z = 0;

  modelstate.twist.angular.x = 0;
  modelstate.twist.angular.y = 0;
  modelstate.twist.angular.z = 0;

}

//----------------------------------------------------------------------------

// set states of objects in gazebo that we spawned

double theta = 0.0;

void motionCallback(const ros::TimerEvent& t)
{
  int i;

  for (i = 0; i < num_objects; i++) {

      modelstate.model_name = object_model_name[i];

      modelstate.pose.position.x = object_x[i];

      if (!(i % 2)) 
	modelstate.pose.position.y = object_y[i] + sin(theta);
      else
	modelstate.pose.position.y = object_y[i] + cos(theta);

      modelstate.pose.position.z = object_z[i];
      
      setmodelstate.request.model_state = modelstate;
      if (!model_state_client.call(setmodelstate))
	printf("set model state %i failed\n", i);
  }

  theta += 0.01;    
}

//----------------------------------------------------------------------------

// returns success if models don't exist (i.e., gazebo has just been started or they were deleted),
// but only seems to SHOW the models if gazebo was just started

void spawn_everything()
{
  int i;

  spawn_model_client = nhp->serviceClient<gazebo_msgs::SpawnModelRequest>("/gazebo/spawn_urdf_model");

  spawnmodel.request.robot_namespace = "gazebo_worlds";
  spawnmodel.request.initial_pose.orientation.x = 0;
  spawnmodel.request.initial_pose.orientation.y = 0;
  spawnmodel.request.initial_pose.orientation.z = 0;
  spawnmodel.request.initial_pose.orientation.w = 1;
  spawnmodel.request.reference_frame = "map";

  for (i = 0; i < num_objects; i++) {

    spawnmodel.request.model_name = object_model_name[i];

    if (!(i % 2)) {
      spawnmodel.request.model_xml = box_urdf;
      object_z.push_back(0.5);
    }
    else {
      spawnmodel.request.model_xml = ball_urdf;
      object_z.push_back(0.25);
    }

    object_x.push_back(3 + 2 * i);
    object_y.push_back(0);

    spawnmodel.request.initial_pose.position.x = object_x[i];
    spawnmodel.request.initial_pose.position.y = object_y[i];
    spawnmodel.request.initial_pose.position.z = object_z[i];

    if (spawn_model_client.call(spawnmodel))
      printf("spawn succeeded\n");
    else
      printf("spawn failed\n");
  }

  //  exit(1);
}

//----------------------------------------------------------------------------

// get rid of models we spawned in gazebo.  they do disappear, but seems that following this
// with spawn_*() doesn't actually make them appear

void delete_everything_and_exit()
{
  int i;

  delete_model_client = nhp->serviceClient<gazebo_msgs::DeleteModelRequest>("/gazebo/delete_model");

  for (i = 0; i < num_objects; i++) {

    deletemodel.request.model_name = object_model_name[i];

    if (delete_model_client.call(deletemodel))
      printf("delete succeeded\n");
    else
      printf("delete failed\n");
  }

  exit(1);
}


//----------------------------------------------------------------------------

int main(int argc, char **argv)
{
  int i;

  // initialize model names

  char str[3];

  for (i = 0; i < num_objects; i++) {
    sprintf(str, "%02d", i); 
    object_model_name.push_back(string("object_") + string(str));
    printf("%i: %s\n", i, object_model_name[i].c_str());
  }

  ros::init(argc, argv, "motion");

  nhp = new ros::NodeHandle();

  for (i = 1; i < argc; i++) {
    if (!strcmp(argv[i], "-spawn"))
      spawn_everything();
    if (!strcmp(argv[i], "-delete"))
      delete_everything_and_exit();
  }

  initialize_modelstates();

  ros::Timer timer = nhp->createTimer(ros::Duration(0.1), motionCallback);

  ros::Rate loop_rate(10);

  while (ros::ok()) {

    ros::spinOnce();
    loop_rate.sleep();

  }

  return 0;
}

//----------------------------------------------------------------------------
//----------------------------------------------------------------------------
